/*
Tiny GLUT program to draw a single red triangle with OpenGL

Orion Sky Lawlor, olawlor@acm.org, 2007/01/15 (Public Domain)
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glew.h> /* GL Extensions Wrangler, http://glew.sourceforge.net/ */
#include <GL/glut.h> /* GL Utilities Toolkit, http://freeglut.sourceforge.net/ */
#include <glew.c> /* avoid linking problems by including GLEW code directly... */

void Exit(const char *where,const char *why="") {
	fprintf (stderr, "FATAL OpenGL Error in %s: %s\n", where, why);
	abort();	
}

// Check GL for errors after the last command
void Check(const char *where) {
	GLenum e=glGetError();
	if (e==GL_NO_ERROR) return;
	const GLubyte *errString = gluErrorString(e);
	Exit(where, (const char *)errString);
}

// Draw a little triangle onscreen
void drawObject(void) {
	glBegin(GL_TRIANGLES);
	glColor4f(1,0.1,0.1,1); /* red */ 
	glNormal3f(0,0,1);
	glVertex3f(-1,-1,0);  
	glVertex3f(-1,+1,0);  
	glVertex3f(+1,+1,0); 
	glEnd();
}

/* This class keeps track of the camera orientation */
class camera_coords {
public:
	double rotX, rotY;
	camera_coords() :rotX(0.0), rotY(0.0) {}
	void nudge(double dx,double dy) {
		rotX+=dx; rotY+=dy;
		glutPostRedisplay(); /* ask to redraw the screen again */
	}
};
camera_coords coords;

/******** Program Initialization *********/
int win_w, win_h; /**< Width and height (in pixels) of our window at last resize */

/* GLUT calls this routine every time our window needs to be redrawn */
void main_display(void) {
	Check("Entering display routine");

// Clear the color and depth buffers
	glClearColor(0.2,0.3,0.4,1.0); /* set clear color */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); /* actually clear color and depth */
	
// Set up GLSL shader
	/* Make a program object */
	GLhandleARB prog=glCreateProgramObjectARB();
        GLint compiled;

	/* Make vertex shader object */
	GLhandleARB vert=glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
	const char *glsl_code=
	"varying vec4 color; \n"
	"varying vec3 normal;\n" //vector pointing away from surface
	"varying vec3 camera;\n" //vector pointing towards camera
	"void main(void) { \n"
	"	color=gl_Color; \n"
	"	normal=normalize(gl_NormalMatrix*gl_Normal); \n"
	"	vec4 world=gl_ModelViewMatrix*gl_Vertex; \n"
	"	camera=normalize(vec3(world)-vec3(0.0,0.0,-2.0)); \n"
	"	gl_Position=gl_ProjectionMatrix*world; \n"
	"}";
	glShaderSourceARB(vert,1,&glsl_code,NULL);
	glCompileShaderARB(vert);
	glAttachObjectARB(prog,vert);
        glGetObjectParameterivARB(vert,GL_OBJECT_COMPILE_STATUS_ARB,&compiled);
        if (!compiled) Exit("Compile error in vertex program\n");

	/* Make fragment shader object */
	GLhandleARB frag=glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
	glsl_code=
	"varying vec4 color;\n"
	"varying vec3 normal;\n"
	"varying vec3 camera; //vector pointing towards camera\n"
	"void main(void) {\n"
	"	vec3 lightDir=normalize(vec3(0,+1,+1));\n"
	"	float light=0.1+0.8*clamp(dot(normal,lightDir),0.0,1.0);\n" // diffuse light
	"	vec3 halfway=normalize(normalize(camera)+lightDir); \n" // blinn's halfway vector
	"	float specular=light*pow(clamp(dot(normal,halfway),0.0,1.0),300.0);\n" // specular light
	"	gl_FragColor=color*light*vec4(1.0,0.0,0.0,1.0)+vec4(specular);\n"
	//"	gl_FragColor=vec4(normal,1.0);\n"
	"}";
	glShaderSourceARB(frag,1,&glsl_code ,NULL);
	glCompileShaderARB(frag);
        glGetObjectParameterivARB(vert,GL_OBJECT_COMPILE_STATUS_ARB,&compiled);
        if (!compiled) Exit("Compile error in fragment program\n");
	glAttachObjectARB(prog,frag);

	/* Link the program. */
	glLinkProgramARB(prog);
        glGetObjectParameterivARB(prog,GL_OBJECT_LINK_STATUS_ARB,&compiled);
        if (!compiled) Exit("Link error between vertex and fragment!\n");

	// Here's how to use the program object for rendering:
	glUseProgramObjectARB(prog);
	
	int loc=glGetUniformLocationARB(prog,"brightness");
	glUniform1fARB(loc,-3.0);
	//printf("Location of brightenss: %d\n",loc);
	
// Set up coordinate system
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(90.0, win_w/(float)win_h, 0.01,100.0); /* set up camera facing down -Z axis */
	glTranslatef(0,0,-2.0); /* push the world origin away from the camera */
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glRotatef(coords.rotX,0,1,0);
	glRotatef(coords.rotY,1,0,0);
	
// Draw the object
	drawObject();
	
// Turn off GLSL
	glUseProgramObjectARB(0);  /* reset back to default fixed-function program */

	// Here's how to delete a program object:
	glDeleteObjectARB(vert);
	glDeleteObjectARB(frag);
	glDeleteObjectARB(prog);
	
// Display the stuff drawn so far to the screen:
	glutSwapBuffers();
	Check("Leaving display routine");
}

/* GLUT calls this routine every time our window changes size */
void main_reshape(int w, int h) {
	win_w=w; win_h=h;
	glViewport(0, 0, (GLsizei)w, (GLsizei)(h));
}

/* GLUT calls this when the mouse moves */
void main_motion(int x,int y) {
	static int last_x=0, last_y=0;
	int dx=x-last_x, dy=y-last_y;
	if (fabs(dx)>30) dx=0; /* silly hack: avoid long jumps */
	if (fabs(dy)>30) dy=0;
	last_x=x; last_y=y;
	
	double scale=1.0; /* degrees rotation / screen pixels mouse motion */
	coords.nudge(scale*dx,scale*dy);
}

int main(int argc, char** argv)
{
	glutInit(&argc,argv);

	// Ask for a back buffer, color buffer, and Z buffer in a window.
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
	glutInitDisplayMode (mode);
	int main_window = glutCreateWindow("Bare-metal OpenGL Demo");
	
	// Turn on OpenGL extensions, like programmable shaders
	//   WARNING: you MUST call glutCreateWindow before doing this!
	//   If you don't call glewInit, or call it wrong, you'll segfault!
	glewInit();

	// Ask GLUT to call us when we need to redraw, reshape, or move
	glutDisplayFunc (main_display);
	glutReshapeFunc (main_reshape);
	glutMotionFunc  (main_motion);
	
	// Process events until user closes our window...
	glutMainLoop();
	
	return 0;
}
